//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/keys.c                                                 $
// $Revision:: 4                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 2/5/02 7:37 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
//

#include "client.h"
#include <ctype.h>

/*
key up events are sent even if in console mode
*/

#define		MAXCMDLINE	256
char	key_lines[32][MAXCMDLINE];
int		key_linepos;
int		shift_down = false;
int		anykeydown;

int		edit_line = 0;
int		history_line = 0;

int		key_waiting;
char	*keybindings[256];
char	*keydoublebindings[256];
char	*keyaltbindings[256];
unsigned keydowntime[256];
qboolean	keydoubledown[256];
qboolean	keyaltdown[256];
cvar_t	*k_doubleclicktime;
cvar_t	*con_new_shortcut_keys; /* FS: Added */
int		alt_down = false;
qboolean	consolekeys[256];	// if true, can't be rebound while in console
qboolean	menubound[256];	// if true, can't be rebound while in menu
int		keyshift[256];		// key to map to if shift held down in console
int		key_repeats[256];	// if > 1, it is autorepeating
qboolean	keydown[256];

typedef struct
{
	const char	*name;
	int		keynum;
} keyname_t;

static keyname_t keynames[] =
{
	{"TAB", K_TAB},
	{"ENTER", K_ENTER},
	{"ESCAPE", K_ESCAPE},
	{"SPACE", K_SPACE},
	{"BACKSPACE", K_BACKSPACE},
	{"UPARROW", K_UPARROW},
	{"DOWNARROW", K_DOWNARROW},
	{"LEFTARROW", K_LEFTARROW},
	{"RIGHTARROW", K_RIGHTARROW},

	{"ALT", K_ALT},
	{"CTRL", K_CTRL},
	{"SHIFT", K_SHIFT},

	{"F1", K_F1},
	{"F2", K_F2},
	{"F3", K_F3},
	{"F4", K_F4},
	{"F5", K_F5},
	{"F6", K_F6},
	{"F7", K_F7},
	{"F8", K_F8},
	{"F9", K_F9},
	{"F10", K_F10},
	{"F11", K_F11},
	{"F12", K_F12},

	{"INS", K_INS},
	{"DEL", K_DEL},
	{"PGDN", K_PGDN},
	{"PGUP", K_PGUP},
	{"HOME", K_HOME},
	{"END", K_END},

	{"MOUSE1", K_MOUSE1},
	{"MOUSE2", K_MOUSE2},
	{"MOUSE3", K_MOUSE3},

	{"JOY1", K_JOY1},
	{"JOY2", K_JOY2},
	{"JOY3", K_JOY3},
	{"JOY4", K_JOY4},

	{"AUX1", K_AUX1},
	{"AUX2", K_AUX2},
	{"AUX3", K_AUX3},
	{"AUX4", K_AUX4},
	{"AUX5", K_AUX5},
	{"AUX6", K_AUX6},
	{"AUX7", K_AUX7},
	{"AUX8", K_AUX8},
	{"AUX9", K_AUX9},
	{"AUX10", K_AUX10},
	{"AUX11", K_AUX11},
	{"AUX12", K_AUX12},
	{"AUX13", K_AUX13},
	{"AUX14", K_AUX14},
	{"AUX15", K_AUX15},
	{"AUX16", K_AUX16},
	{"AUX17", K_AUX17},
	{"AUX18", K_AUX18},
	{"AUX19", K_AUX19},
	{"AUX20", K_AUX20},
	{"AUX21", K_AUX21},
	{"AUX22", K_AUX22},
	{"AUX23", K_AUX23},
	{"AUX24", K_AUX24},
	{"AUX25", K_AUX25},
	{"AUX26", K_AUX26},
	{"AUX27", K_AUX27},
	{"AUX28", K_AUX28},
	{"AUX29", K_AUX29},
	{"AUX30", K_AUX30},
	{"AUX31", K_AUX31},
	{"AUX32", K_AUX32},

	{"KP_HOME", K_KP_HOME },
	{"KP_UPARROW", K_KP_UPARROW },
	{"KP_PGUP", K_KP_PGUP },
	{"KP_LEFTARROW", K_KP_LEFTARROW },
	{"KP_5", K_KP_5 },
	{"KP_RIGHTARROW", K_KP_RIGHTARROW },
	{"KP_END", K_KP_END },
	{"KP_DOWNARROW", K_KP_DOWNARROW },
	{"KP_PGDN", K_KP_PGDN },
	{"KP_ENTER", K_KP_ENTER },
	{"KP_INS", K_KP_INS },
	{"KP_DEL", K_KP_DEL },
	{"KP_SLASH", K_KP_SLASH },
	{"KP_MINUS", K_KP_MINUS },
	{"KP_PLUS", K_KP_PLUS },

	{"MWHEELUP", K_MWHEELUP },
	{"MWHEELDOWN", K_MWHEELDOWN },
	{"MOUSE4", K_MOUSE4},
	{"MOUSE5", K_MOUSE5},

	{"PAUSE", K_PAUSE},

	{"SEMICOLON", ';'},	// because a raw semicolon seperates commands

	{NULL,0}
};

/*
==============================================================================

			LINE TYPING INTO THE CONSOLE

==============================================================================
*/

static qboolean CheckForCommand (void)
{
	char	command[128];
	const char	*cmd, *s;
	int		i;

	s = key_lines[edit_line]+1;

	for (i = 0; i < 127; i++)
		if (s[i] <= ' ')
			break;
		else
			command[i] = s[i];
	command[i] = 0;

	cmd = Cmd_CompleteCommand (command);
	if (!cmd || strcmp(cmd, command))
		cmd = Cvar_CompleteVariable (command);
	if (!cmd || strcmp(cmd, command))
		return false;		// just a chat message
	return true;
}

static void CompleteCommand (void)
{
	const char 	*cmd;
	char  *s;

	s = key_lines[edit_line]+1;
	if (*s == '\\' || *s == '/')
		s++;

#if 0
	cmd = Cmd_CompleteCommand (s);
	if (!cmd)
		cmd = Cvar_CompleteVariable (s);
#else
	cmd = Sort_Possible_Cmds(s); /* FS: New autocomplete */
#endif

	if (cmd)
	{
		key_lines[edit_line][1] = '/';
		strcpy (key_lines[edit_line]+2, cmd);
		key_linepos = (int)strlen(cmd)+2;
		key_lines[edit_line][key_linepos] = ' ';
		key_linepos++;
		key_lines[edit_line][key_linepos] = 0;
		return;
	}
}

/*
====================
Key_Console

Interactive line editing and console scrollback
====================
*/
static void Key_Console (int key)
{
	switch (key)
	{
	case K_KP_SLASH:
		key = '/';
		break;
	case K_KP_MINUS:
		key = '-';
		break;
	case K_KP_PLUS:
		key = '+';
		break;
	case K_KP_HOME:
		key = '7';
		break;
	case K_KP_UPARROW:
		key = '8';
		break;
	case K_KP_PGUP:
		key = '9';
		break;
	case K_KP_LEFTARROW:
		key = '4';
		break;
	case K_KP_5:
		key = '5';
		break;
	case K_KP_RIGHTARROW:
		key = '6';
		break;
	case K_KP_END:
		key = '1';
		break;
	case K_KP_DOWNARROW:
		key = '2';
		break;
	case K_KP_PGDN:
		key = '3';
		break;
	case K_KP_INS:
		key = '0';
		break;
	case K_KP_DEL:
		key = '.';
		break;
	}

	if ((toupper(key) == 'V' && keydown[K_CTRL]) ||
	    ((key == K_INS || key == K_KP_INS) && keydown[K_SHIFT]) )
	{
		char *cbd;
		if ((cbd = Sys_GetClipboardData()) != NULL)
		{
			int i;

			strtok(cbd, "\n\r\b");

			i = (int)strlen(cbd);
			if (i + key_linepos >= MAXCMDLINE)
				i= MAXCMDLINE - key_linepos;

			if (i > 0)
			{
				cbd[i] = 0;
				strcat(key_lines[edit_line], cbd);
				key_linepos += i;
			}
			free(cbd);
		}

		return;
	}

	if (con_new_shortcut_keys->intValue) /* FS: Added */
	{
		if (key == 'c' && keydown[K_CTRL])
		{
#ifdef GAMESPY
			if (cls.gamespyupdate) {
				Cbuf_AddText("gspystop\n");
				return;
			}
#endif
			Cbuf_AddText ("disconnect\n");
			return;
		}

		if (key == 'r' && keydown[K_CTRL])
		{
			Cbuf_AddText ("reconnect\n");
			return;
		}
	}

	if (key == 'l' && keydown[K_CTRL])
	{
		Cbuf_AddText ("clear\n");
		return;
	}

	if (key == K_ENTER || key == K_KP_ENTER)
	{	// backslash text are commands, else chat
		if (key_lines[edit_line][1] == '\\' || key_lines[edit_line][1] == '/')
		{
			Cbuf_AddText (key_lines[edit_line]+2);	// skip the >
		}
		else if (CheckForCommand())
		{
			Cbuf_AddText (key_lines[edit_line]+1);	// valid command
		}
		else
		{	// convert to a chat message
			Cbuf_AddText (key_lines[edit_line]+1);	// skip the >
		}

		Cbuf_AddText ("\n");
		Com_Printf ("%s\n",key_lines[edit_line]);
		edit_line = (edit_line + 1) & 31;
		history_line = edit_line;
		key_lines[edit_line][0] = ']';
		key_linepos = 1;
		if (cls.state == ca_disconnected)
			SCR_UpdateScreen ();	// force an update, because the command
									// may take some time

		Sort_Possible_Cmds_Cleanup(); /* FS: New autocomplete */

		return;
	}

	if (key == K_TAB)
	{	// command completion
		CompleteCommand ();
		return;
	}

	if (key == K_BACKSPACE || key == K_LEFTARROW || key == K_KP_LEFTARROW || (key == 'h' && keydown[K_CTRL]))
	{
		if (key_linepos > 1)
			key_linepos--;
		return;
	}

	if (key == K_UPARROW || key == K_KP_UPARROW || (key == 'p' && keydown[K_CTRL]))
	{
		do
		{
			history_line = (history_line - 1) & 31;
		} while (history_line != edit_line
				&& !key_lines[history_line][1]);
		if (history_line == edit_line)
			history_line = (edit_line+1)&31;
		strcpy(key_lines[edit_line], key_lines[history_line]);
		key_linepos = (int)strlen(key_lines[edit_line]);
		return;
	}

	if (key == K_DOWNARROW || key == K_KP_DOWNARROW || (key == 'n' && keydown[K_CTRL]))
	{
		if (history_line == edit_line) return;
		do
		{
			history_line = (history_line + 1) & 31;
		}
		while (history_line != edit_line
			&& !key_lines[history_line][1]);
		if (history_line == edit_line)
		{
			key_lines[edit_line][0] = ']';
			key_linepos = 1;
		}
		else
		{
			strcpy(key_lines[edit_line], key_lines[history_line]);
			key_linepos = (int)strlen(key_lines[edit_line]);
		}
		return;
	}

	if (key == K_PGUP || key == K_KP_PGUP || key == K_MWHEELUP) /* FS: Added scroll wheel */
	{
		con.display -= 2;
		return;
	}

	if (key == K_PGDN || key == K_KP_PGDN || key == K_MWHEELDOWN) /* FS: Added scroll wheel */
	{
		con.display += 2;
		if (con.display > con.current)
			con.display = con.current;
		return;
	}

	if (key == K_HOME || key == K_KP_HOME)
	{
		con.display = con.current - con.totallines + 10;
		return;
	}

	if (key == K_END || key == K_KP_END)
	{
		con.display = con.current;
		return;
	}

	if (key < 32 || key > 127)
		return;	// non printable

	if (key_linepos < MAXCMDLINE-1)
	{
		key_lines[edit_line][key_linepos] = key;
		key_linepos++;
		key_lines[edit_line][key_linepos] = 0;
	}

}

//============================================================================

#define MAX_CHAT_BUFFERS 32

qboolean	chat_team;
char		chat_buffer[MAXCMDLINE];
int			chat_bufferlen = 0;
qboolean	chat_talk;

char		chat_buffer_array[MAX_CHAT_BUFFERS][MAXCMDLINE]; /* FS: Chat history */
int			chat_head = 0, chat_tail = 0; /* FS: Chat history */
int			chat_index = 0; /* FS: Chat history */

static void Key_Message (int key)
{
	if (key == K_ENTER || key == K_KP_ENTER)
	{
		if (chat_talk)
			Cbuf_AddText ("say_talk \"");
		else if (chat_team)
			Cbuf_AddText ("say_team \"");
		else
			Cbuf_AddText ("say \"");

		Cbuf_AddText(chat_buffer);
		Cbuf_AddText("\"\n");

		cls.key_dest = key_game;

		/* FS: Chat history */
		strcpy (chat_buffer_array[chat_head], chat_buffer);
		chat_head = (chat_head + 1) % MAX_CHAT_BUFFERS;
		if (chat_head == chat_tail)
			chat_tail = (chat_tail + 1) % MAX_CHAT_BUFFERS;
		chat_buffer_array[chat_head][0] = 0;
		chat_index = chat_head;

		chat_bufferlen = 0;
		chat_buffer[0] = 0;

		return;
	}

	if (key == K_ESCAPE)
	{
		cls.key_dest = key_game;
		chat_index = chat_head; /* FS: Reset to the beginning of the chat history */
		chat_bufferlen = 0;
		chat_buffer[0] = 0;
		return;
	}

	if (key == K_UPARROW || key == K_RIGHTARROW) /* FS: Press up to cycle up the index to the first chat msg */
	{
		cls.key_dest = key_message;

		if (chat_index != chat_tail)
		{
			chat_index = (chat_index + MAX_CHAT_BUFFERS - 1) % MAX_CHAT_BUFFERS;
			strcpy (chat_buffer, chat_buffer_array[chat_index]);
		}

		chat_bufferlen = (int)(strlen(chat_buffer));
		return;
	}

	if (key == K_DOWNARROW || key == K_LEFTARROW) /* FS: Press down to cycle down the index to the last chat msg */
	{
		cls.key_dest = key_message;

		if (chat_index != chat_head) {
			chat_index = (chat_index + 1) % MAX_CHAT_BUFFERS;
			strcpy (chat_buffer, chat_buffer_array[chat_index]);
		}

		chat_bufferlen = (int)(strlen(chat_buffer));
		return;
	}

	if (key < 32 || key > 127)
		return;	// non printable

	if (key == K_BACKSPACE)
	{
		if (chat_bufferlen)
		{
			chat_bufferlen--;
			chat_buffer[chat_bufferlen] = 0;
		}
		return;
	}

	if (chat_bufferlen == sizeof(chat_buffer)-1)
		return; // all full

	chat_buffer[chat_bufferlen++] = key;
	chat_buffer[chat_bufferlen] = 0;
}


static void Key_Menu3D (int key, unsigned time)
{
	console3d_t *p = cls.currentconsole;
	item3d_t    *item;

	if (!p)
		return;

	if (!p->menu3d || !p->menu3d[p->current_menu_level])
	{
		CONSOLE_Exit(p);
		return;
	}

	item = p->menu3d[p->current_menu_level]->item[p->sel_menu_item];

	// Only do help messages on the main menu.
	if (p == cls.mainconsole && key == K_F1)
	{
		if (item && item->helpdata)
			CONSOLE_ScreenPrint(item->helpdata, item->helptime);
		return;
	}

	if (p->menu3d[p->current_menu_level]->keyfunc)
	{
		p->menu3d[p->current_menu_level]->keyfunc(key);
		return;
	}

	if (item && item->type == BIND3D)
	{
		MENU_Bind_Key_f(&item->bind, key, time);
		return;
	}

	if (item && item->type == FIELD3D)
	{
		if (MENU_Field_Key_f(&item->field, key))
			return;
	}

	if (key == K_ESCAPE)
	{
		qboolean ret;
		ret = CONSOLE_PopMenu(p);
		if (!ret)
		{
			CL_RandomLocalSound(menu_out_sound);
			CONSOLE_Exit(p);
		}
	}

	if (key == K_ENTER ||  key == K_KP_ENTER)
	{
		if (item)
			CONSOLE_ProcessMenuAction(p, item->action);

		// If the user presses return on the field, then advance the cursor down
		if (item && item->type == FIELD3D)
			CONSOLE_SelectItem(p, 1);

		return;
	}

	if (key == K_RIGHTARROW || key == K_KP_RIGHTARROW)
	{
		if (item)
			CONSOLE_SlideItem(p, 1, item);
		return;
	}

	if (key == K_LEFTARROW || key == K_KP_LEFTARROW)
	{
		if (item)
			CONSOLE_SlideItem(p, -1, item);
		return;
	}

	if (key == K_UPARROW || key == K_KP_UPARROW || (key == 'p' && keydown[K_CTRL]))
	{
		CONSOLE_SelectItem(p, -1);
		return;
	}

	if (key == K_DOWNARROW || key == K_KP_DOWNARROW || (key == 'n' && keydown[K_CTRL]))
	{
		CONSOLE_SelectItem(p, 1);
		return;
	}
}


static void Key_Console3D (int key)
{
	console3d_t *p = cls.currentconsole;

#if 0
	if (key == K_TAB)
	{
		if (p->menucon)
		{
			p->focus = MENU3D;
			cls.key_dest = key_menu3d;
		}
	}
#endif

	if (key == K_ENTER ||  key == K_KP_ENTER)
	{
		CONSOLE_SendCharacterToServer(p, key);
		return;
	}

	if (key == K_BACKSPACE || key == K_LEFTARROW || (key == 'h' && keydown[K_CTRL]))
	{
		if (p->key_linepos > 1)
			CONSOLE_SendCharacterToServer(p, -1);
		return;
	}

#if 0
	if (key == K_UPARROW || (key == 'p' && keydown[K_CTRL]))
	{
		do
		{
			p->history_line = (p->history_line - 1) & 31;
		} while (p->history_line != p->edit_line && !p->key_lines[p->history_line][1]);

		if (p->history_line == p->edit_line)
			p->history_line = (p->edit_line+1)&31;

		strcpy(p->key_lines[p->edit_line], p->key_lines[p->history_line]);
		p->key_linepos = strlen(p->key_lines[p->edit_line]);
		return;
	}

	if (key == K_DOWNARROW || (key == 'n' && keydown[K_CTRL]))
	{
		if (p->history_line == p->edit_line)
			return;

		do
		{
			p->history_line = (p->history_line + 1) & 31;
		}
		while (p->history_line != p->edit_line	&& !p->key_lines[p->history_line][1]);

		if (p->history_line == p->edit_line)
		{
			p->key_lines[p->edit_line][0] = ']';
			p->key_linepos = 1;
		}
		else
		{
			strcpy(p->key_lines[p->edit_line], p->key_lines[p->history_line]);
			p->key_linepos = strlen(p->key_lines[p->edit_line]);
		}

		return;
	}
#endif

	if (key == K_PGUP || key == K_KP_PGUP)
	{
		p->con.display -= 2;
		return;
	}

	if (key == K_PGDN || key == K_KP_PGDN)
	{
		p->con.display += 2;
		if (p->con.display > p->con.current)
			p->con.display = p->con.current;
		return;
	}

	if (key == K_HOME || key == K_KP_HOME)
	{
		p->con.display = p->con.current - p->con.totallines + 10;
		return;
	}

	if (key == K_END || key == K_KP_END)
	{
		p->con.display = p->con.current;
		return;
	}

	if (key == K_ESCAPE)
	{
		CONSOLE_Exit(p);
		return;
	}

	if (key < 32 || key > 127)
		return;	// non printable

	if (p->key_linepos < MAXCMDLINE-1)
	{
		CONSOLE_SendCharacterToServer(p,key);
	}
}

//============================================================================


/*
===================
Key_StringToKeynum

Returns a key number to be used to index keybindings[] by looking at
the given string.  Single ascii characters return themselves, while
the K_* names are matched up.
===================
*/
static int Key_StringToKeynum (const char *str)
{
	keyname_t	*kn;

	if (!str || !str[0])
		return -1;
	if (!str[1])
		return str[0];

	for (kn = keynames; kn->name; kn++)
	{
		if (!Q_stricmp(str,kn->name))
			return kn->keynum;
	}
	return -1;
}

/*
===================
Key_KeynumToString

Returns a string (either a single ascii char, or a K_* name) for the
given keynum.
FIXME: handle quote special (general escape sequence?)
===================
*/
const char *Key_KeynumToString (int keynum)
{
	keyname_t	*kn;
	static	char	tinystr[2];

	if (keynum == -1)
		return "<KEY NOT FOUND>";
	if (keynum > 32 && keynum < 127)
	{	// printable ascii
		tinystr[0] = keynum;
		tinystr[1] = 0;
		return tinystr;
	}

	for (kn = keynames; kn->name; kn++)
		if (keynum == kn->keynum)
			return kn->name;

	return "<UNKNOWN KEYNUM>";
}

void Key_ReleaseBindings(void)
{
	int i;
	i = 0;
	do
	{
		free(keybindings[i]);
		keybindings[i] = 0;

		free(keydoublebindings[i]);
		keydoublebindings[i] = 0;

		free(keyaltbindings[i]);
		keyaltbindings[i] = 0;
	} while (++i < 256);
}

/*
===================
Key_SetBinding
===================
*/
void Key_SetBinding (int keynum, const char *binding)
{
	if (keynum == -1)
		return;

// free old bindings for this key
	free(keybindings[keynum]);

// allocate memory for new binding
	keybindings[keynum] = strdup(binding);
}

/*
===================
Key_SetDoubleBinding
===================
*/
void Key_SetDoubleBinding (int keynum, const char *binding)
{
	if (keynum == -1)
		return;

// free old bindings
	free(keydoublebindings[keynum]);

// allocate memory for new binding
	keydoublebindings[keynum] = strdup(binding);
}

/*
===================
Key_Undoublebind_f
===================
*/
static void Key_Undoublebind_f (void)
{
	int		b;

	if (Cmd_Argc() != 2)
	{
		Com_Printf ("undoublebind <key> : remove commands from a key\n");
		return;
	}

	b = Key_StringToKeynum (Cmd_Argv(1));
	if (b == -1)
	{
		Com_Printf ("\"%s\" isn't a valid key\n", Cmd_Argv(1));
		return;
	}

	Key_SetDoubleBinding (b, "");
}


/*
===================
Key_DoubleBind_f
===================
*/
static void Key_DoubleBind_f (void)
{
	int			i, c, b;
	char		cmd[1024];

	c = Cmd_Argc();
	if (c < 2)
	{
		Com_Printf ("doublebind <key> [command] : attach a command to a key\n");
		return;
	}

	b = Key_StringToKeynum (Cmd_Argv(1));
	if (b == -1)
	{
		Com_Printf ("\"%s\" isn't a valid key\n", Cmd_Argv(1));
		return;
	}

	if (c == 2)
	{
		if (keydoublebindings[b])
			Com_Printf ("\"%s\" = \"%s\"\n", Cmd_Argv(1), keydoublebindings[b]);
		else
			Com_Printf ("\"%s\" is not double bound\n", Cmd_Argv(1));
		return;
	}

// copy the rest of the command line
	cmd[0] = 0;		// start out with a null string
	for (i = 2; i < c; i++)
	{
		strcat (cmd, Cmd_Argv(i));
		if (i != (c-1))
			strcat (cmd, " ");
	}

	Key_SetDoubleBinding (b, cmd);
}

/*
===================
Key_SetAltBinding
===================
*/
static void Key_SetAltBinding (int keynum, const char *binding)
{
	if (keynum == -1)
		return;

// free old bindings
	free(keyaltbindings[keynum]);

// allocate memory for new binding
	keyaltbindings[keynum] = strdup(binding);
}

/*
===================
Key_Unaltbind_f
===================
*/
static void Key_Unaltbind_f (void)
{
	int		b;

	if (Cmd_Argc() != 2)
	{
		Com_Printf ("unaltbind <key> : remove commands from a key\n");
		return;
	}

	b = Key_StringToKeynum (Cmd_Argv(1));
	if (b == -1)
	{
		Com_Printf ("\"%s\" isn't a valid key\n", Cmd_Argv(1));
		return;
	}

	Key_SetAltBinding (b, "");
}


/*
===================
Key_AltBind_f
===================
*/
static void Key_AltBind_f (void)
{
	int			i, c, b;
	char		cmd[1024];

	c = Cmd_Argc();

	if (c < 2)
	{
		Com_Printf ("altbind <key> [command] : attach a command to a key\n");
		return;
	}
	b = Key_StringToKeynum (Cmd_Argv(1));
	if (b == -1)
	{
		Com_Printf ("\"%s\" isn't a valid key\n", Cmd_Argv(1));
		return;
	}

	if (c == 2)
	{
		if (keyaltbindings[b])
			Com_Printf ("\"%s\" = \"%s\"\n", Cmd_Argv(1), keyaltbindings[b]);
		else
			Com_Printf ("\"%s\" is not alt-bound\n", Cmd_Argv(1));
		return;
	}

// copy the rest of the command line
	cmd[0] = 0;		// start out with a null string
	for (i = 2; i < c; i++)
	{
		strcat (cmd, Cmd_Argv(i));
		if (i != (c-1))
			strcat (cmd, " ");
	}

	Key_SetAltBinding (b, cmd);
}

/*
===================
Key_Unbind_f
===================
*/
static void Key_Unbind_f (void)
{
	int		b;

	if (Cmd_Argc() != 2)
	{
		Com_Printf ("unbind <key> : remove commands from a key\n");
		return;
	}

	b = Key_StringToKeynum (Cmd_Argv(1));
	if (b == -1)
	{
		Com_Printf ("\"%s\" isn't a valid key\n", Cmd_Argv(1));
		return;
	}

	Key_SetBinding (b, "");
}

static void Key_Unbindall_f (void)
{
	int		i;

	for (i = 0; i < 256; i++)
		if (keybindings[i])
			Key_SetBinding (i, "");
}


/*
===================
Key_Bind_f
===================
*/
static void Key_Bind_f (void)
{
	int			i, c, b;
	char		cmd[1024];

	c = Cmd_Argc();

	if (c < 2)
	{
		Com_Printf ("bind <key> [command] : attach a command to a key\n");
		return;
	}
	b = Key_StringToKeynum (Cmd_Argv(1));
	if (b == -1)
	{
		Com_Printf ("\"%s\" isn't a valid key\n", Cmd_Argv(1));
		return;
	}

	if (c == 2)
	{
		if (keybindings[b])
			Com_Printf ("\"%s\" = \"%s\"\n", Cmd_Argv(1), keybindings[b]);
		else
			Com_Printf ("\"%s\" is not bound\n", Cmd_Argv(1));
		return;
	}

// copy the rest of the command line
	cmd[0] = 0;		// start out with a null string
	for (i = 2 ; i < c; i++)
	{
		strcat (cmd, Cmd_Argv(i));
		if (i != (c-1))
			strcat (cmd, " ");
	}

	Key_SetBinding (b, cmd);
}

/*
============
Key_WriteBindings

Writes lines containing "bind key value"
============
*/
void Key_WriteBindings (FILE *f)
{
	int		i;

	for (i = 0; i < 256; i++)
	{
		if (keybindings[i] && keybindings[i][0])
			fprintf (f, "bind %s \"%s\"\n", Key_KeynumToString(i), keybindings[i]);
		if (keydoublebindings[i] && keydoublebindings[i][0])
			fprintf (f, "doublebind %s \"%s\"\n", Key_KeynumToString(i), keydoublebindings[i]);
		if (keyaltbindings[i] && keyaltbindings[i][0])
			fprintf (f, "altbind %s \"%s\"\n", Key_KeynumToString(i), keyaltbindings[i]);
	}
}


/*
============
Key_Bindlist_f

============
*/
static void Key_Bindlist_f (void)
{
	int		i;

	Com_Printf ("Single Bindings\n");
	for (i = 0; i < 256; i++)
		if (keybindings[i] && keybindings[i][0])
			Com_Printf ("%s \"%s\"\n", Key_KeynumToString(i), keybindings[i]);
}

/*
============
Key_DoubleBindlist_f

============
*/
static void Key_DoubleBindlist_f (void)
{
	int i;

	Com_Printf ("Double Bindings\n");
	for (i = 0; i < 256; i++)
		if (keydoublebindings[i] && keydoublebindings[i][0])
			Com_Printf ("%s \"%s\"\n", Key_KeynumToString(i), keydoublebindings[i]);
}

/*
============
Key_AltBindlist_f

============
*/
static void Key_AltBindlist_f (void)
{
	int i;

	Com_Printf ("Alt Bindings\n");
	for (i = 0; i < 256; i++)
		if (keyaltbindings[i] && keyaltbindings[i][0])
			Com_Printf ("%s \"%s\"\n", Key_KeynumToString(i), keyaltbindings[i]);
}

/*
===================
Key_Init
===================
*/
void Key_Init (void)
{
	int		i;

	for (i = 0; i < 32; i++)
	{
		key_lines[i][0] = ']';
		key_lines[i][1] = 0;
	}
	key_linepos = 1;

//
// init ascii characters in console mode
//
	for (i = 32; i < 128; i++)
		consolekeys[i] = true;

	consolekeys[K_ENTER] = true;
	consolekeys[K_KP_ENTER] = true;
	consolekeys[K_TAB] = true;
	consolekeys[K_LEFTARROW] = true;
	consolekeys[K_KP_LEFTARROW] = true;
	consolekeys[K_RIGHTARROW] = true;
	consolekeys[K_KP_RIGHTARROW] = true;
	consolekeys[K_UPARROW] = true;
	consolekeys[K_KP_UPARROW] = true;
	consolekeys[K_DOWNARROW] = true;
	consolekeys[K_KP_DOWNARROW] = true;
	consolekeys[K_BACKSPACE] = true;
	consolekeys[K_HOME] = true;
	consolekeys[K_KP_HOME] = true;
	consolekeys[K_END] = true;
	consolekeys[K_KP_END] = true;
	consolekeys[K_PGUP] = true;
	consolekeys[K_KP_PGUP] = true;
	consolekeys[K_PGDN] = true;
	consolekeys[K_KP_PGDN] = true;
	consolekeys[K_SHIFT] = true;
	consolekeys[K_INS] = true;
	consolekeys[K_KP_INS] = true;
	consolekeys[K_KP_DEL] = true;
	consolekeys[K_KP_SLASH] = true;
	consolekeys[K_KP_PLUS] = true;
	consolekeys[K_KP_MINUS] = true;
	consolekeys[K_KP_5] = true;
	consolekeys['`'] = false;
	consolekeys['~'] = false;

	for (i = 0; i < 256; i++)
		keyshift[i] = i;
	for (i = 'a'; i <= 'z'; i++)
		keyshift[i] = i - 'a' + 'A';
	keyshift['1'] = '!';
	keyshift['2'] = '@';
	keyshift['3'] = '#';
	keyshift['4'] = '$';
	keyshift['5'] = '%';
	keyshift['6'] = '^';
	keyshift['7'] = '&';
	keyshift['8'] = '*';
	keyshift['9'] = '(';
	keyshift['0'] = ')';
	keyshift['-'] = '_';
	keyshift['='] = '+';
	keyshift[','] = '<';
	keyshift['.'] = '>';
	keyshift['/'] = '?';
	keyshift[';'] = ':';
	keyshift['\''] = '"';
	keyshift['['] = '{';
	keyshift[']'] = '}';
	keyshift['`'] = '~';
	keyshift['\\'] = '|';

	menubound[K_ESCAPE] = true;
	for (i = 0; i < 12; i++)
		menubound[K_F2+i] = true;

//
// register our functions
//
	Cmd_AddCommand ("bind",Key_Bind_f);
	Cmd_AddCommand ("unbind",Key_Unbind_f);
	Cmd_AddCommand ("unbindall",Key_Unbindall_f);
	Cmd_AddCommand ("bindlist",Key_Bindlist_f);
	Cmd_AddCommand ("doublebind",Key_DoubleBind_f);
	Cmd_AddCommand ("undoublebind",Key_Undoublebind_f);
	Cmd_AddCommand ("doublebindlist",Key_DoubleBindlist_f);
	memset(keydowntime, 0, sizeof(keydowntime));
	memset(keydoubledown, 0, sizeof(keydoubledown));
	k_doubleclicktime = Cvar_Get ("k_doubleclicktime", "250", CVAR_ARCHIVE);

	/* FS: Added */
	con_new_shortcut_keys = Cvar_Get ("con_new_shortcut_keys", "0", CVAR_ARCHIVE);
	con_new_shortcut_keys->description = "Enables CTRL+C, CTRL+R, CTRL+Q, and ALT+F4 shortcut keys.";

	Cmd_AddCommand ("altbind",Key_AltBind_f);
	Cmd_AddCommand ("unaltbind",Key_Unaltbind_f);
	Cmd_AddCommand ("altbindlist",Key_AltBindlist_f);
	memset(keyaltdown, 0, sizeof(keyaltdown));
}

/*
===================
Key_Event

Called by the system between frames for both key up and key down events
Should NOT be called during an interrupt!
===================
*/
void Key_Event (int key, qboolean down, unsigned time)
{
	char	*kb;
	char	cmd[1024];

	// hack for modal presses
	if (key_waiting == -1)
	{
		if (down)
			key_waiting = key;
		return;
	}

	// update auto-repeat status
	if (down)
	{
		key_repeats[key]++;

		if (!cl_autorepeat_allkeys->intValue) /* FS: Added */
		{

			if (key != K_BACKSPACE
				&& key != K_PAUSE
				&& key != K_PGUP
				&& key != K_KP_PGUP
				&& key != K_PGDN
				&& key != K_KP_PGDN
//HACK HACK
//FIXME FIXME
				&& key != '['
				&& key != ']'
				&& key_repeats[key] > 1)
			{
				return;	// ignore most autorepeats
			}
		}

		if (key >= 200 && !keybindings[key] && !keydoublebindings[key])
			Com_Printf ("%s is unbound, hit F4 to set.\n", Key_KeynumToString(key));
		//
		// check for double-click
		//
		if ((time - keydowntime[key]) <= k_doubleclicktime->value)
			keydoubledown[key] = down;
		keydowntime[key] = time;
	}
	else
	{
		key_repeats[key] = 0;
	}

	if (key == K_SHIFT)
		shift_down = down;

	if (key == K_ALT)
		alt_down = down;

	// console key is hardcoded, so the user can never unbind it
	if (key == '`' || key == '~')
	{
		if (!down)
			return;
		Con_ToggleConsole_f ();
		return;
	}

	if (con_new_shortcut_keys->intValue) /* FS: Added */
	{
		if (key == 'q' && keydown[K_CTRL])
		{
			Cbuf_AddText ("disconnect ; quit\n");
			return;
		}

		if (keydown [K_F4] && keydown[K_ALT])
		{
			Cbuf_AddText ("disconnect ; quit\n");
			return;
		}
	}

	// any key during the attract mode will bring up the menu
	if (cl.attractloop && cls.key_dest != key_menu && cls.key_dest != key_menu3d && key < 200)
	{
		SCR_StopCinematic();
	}
	// menu key is hardcoded, so the user can never unbind it
	if (key == K_ESCAPE)
	{
		if (!down)
			return;

		switch (cls.key_dest)
		{
		case key_message:
			Key_Message (key);
			break;
		case key_menu:
			Com_Error (ERR_FATAL, "Bad cls.key_dest");
			break;
		case key_game:
		case key_console:
			MENU_Main_f ();
			break;
		case key_console3d:
			Key_Console3D(key);
			break;
		case key_menu3d:
			Key_Menu3D(key, time);
			break;
		default:
			Com_Error (ERR_FATAL, "Bad cls.key_dest");
		}
		return;
	}

	// track if any key is down for BUTTON_ANY
	if (down)
		keydown[key] = down;

	if (down)
	{
		if (key_repeats[key] == 1)
			anykeydown++;
	}
	else
	{
		anykeydown--;
		if (anykeydown < 0)
			anykeydown = 0;
	}

	// see if it was an ALT key
	if (down && alt_down && key != K_ALT)
	{
		keyaltdown[key] = down;
	}

//
// key up events only generate commands if the game key binding is
// a button command (leading + sign).  These will occur even in console mode,
// to keep the character from continuing an action started before a console
// switch.  Button commands include the kenum as a parameter, so multiple
// downs can be matched with ups
//
	if (!down)
	{
		if (keyaltdown[key] || keyaltdown[keyshift[key]])
		{
			keyaltdown[key] = 0;
			kb = keyaltbindings[key];
			if (kb && kb[0] == '+')
			{
				Com_sprintf (cmd, sizeof(cmd), "-%s %i %i\n", kb+1, key, time);
				Cbuf_AddText (cmd);
			}
			if (keyshift[key] != key)
			{
				kb = keyaltbindings[keyshift[key]];
				if (kb && kb[0] == '+')
				{
					Com_sprintf (cmd, sizeof(cmd), "-%s %i %i\n", kb+1, key, time);
					Cbuf_AddText (cmd);
				}
			}
		}

		if (keydown[key] || keydown[keyshift[key]])
		{
			keydown[key] = 0;
			kb = keybindings[key];
			if (kb && kb[0] == '+')
			{
				Com_sprintf (cmd, sizeof(cmd), "-%s %i %i\n", kb+1, key, time);
				Cbuf_AddText (cmd);
			}
			if (keyshift[key] != key)
			{
				kb = keybindings[keyshift[key]];
				if (kb && kb[0] == '+')
				{
					Com_sprintf (cmd, sizeof(cmd), "-%s %i %i\n", kb+1, key, time);
					Cbuf_AddText (cmd);
				}
			}
		}

		if (keydoubledown[key] || keydoubledown[keyshift[key]])
		{
			keydoubledown[key] = 0;
			kb = keydoublebindings[key];
			if (kb && kb[0] == '+')
			{
				Com_sprintf (cmd, sizeof(cmd), "-%s %i %i\n", kb+1, key, time);
				Cbuf_AddText (cmd);
			}
			if (keyshift[key] != key)
			{
				kb = keydoublebindings[keyshift[key]];
				if (kb && kb[0] == '+')
				{
					Com_sprintf (cmd, sizeof(cmd), "-%s %i %i\n", kb+1, key, time);
					Cbuf_AddText (cmd);
				}
			}
		}

		return;
	}

//
// if not a consolekey, send to the interpreter no matter what mode is
//

	if ((cls.key_dest == key_console && !consolekeys[key]) ||
	    (cls.key_dest == key_console3d && menubound[key]) ||
	    (cls.key_dest == key_menu3d && menubound[key]) ||
	    (cls.key_dest == key_game && (cls.state == ca_active || !consolekeys[key])) )
	{
		kb = keyaltbindings[key];
		if (kb && keyaltdown[key])
		{
			if (kb[0] == '+')
			{	// button commands add keynum and time as a parm
				Com_sprintf (cmd, sizeof(cmd), "%s %i %i\n", kb, key, time);
				Cbuf_AddText (cmd);
			}
			else
			{
				Cbuf_AddText (kb);
				Cbuf_AddText ("\n");
			}
			return;
		}
		kb = keybindings[key];
		if (kb)
		{
			if (kb[0] == '+')
			{	// button commands add keynum and time as a parm
				Com_sprintf (cmd, sizeof(cmd), "%s %i %i\n", kb, key, time);
				Cbuf_AddText (cmd);
			}
			else
			{
				Cbuf_AddText (kb);
				Cbuf_AddText ("\n");
			}
		}
		kb = keydoublebindings[key];
		if (kb && keydoubledown[key])
		{
			if (kb[0] == '+')
			{	// button commands add keynum and time as a parm
				Com_sprintf (cmd, sizeof(cmd), "%s %i %i\n", kb, key, time);
				Cbuf_AddText (cmd);
			}
			else
			{
				Cbuf_AddText (kb);
				Cbuf_AddText ("\n");
			}
		}
		return;
	}

	if (!down)
		return;		// other systems only care about key down events

	if (shift_down)
		key = keyshift[key];

	switch (cls.key_dest)
	{
	case key_message:
		Key_Message (key);
		break;

	case key_menu:
		Com_Error (ERR_FATAL, "Bad cls.key_dest");
		break;

	case key_game:
	case key_console:
		Key_Console (key);
		break;

	case key_console3d:
		Key_Console3D(key);
		break;

	case key_menu3d:
		Key_Menu3D(key, time);
		break;
	default:
		Com_Error (ERR_FATAL, "Bad cls.key_dest");
	}
}

/*
===================
Key_ClearStates
===================
*/
void Key_ClearStates (void)
{
	int		i;

	anykeydown = false;

	for (i = 0; i < 256; i++)
	{
		if (keydown[i] || key_repeats[i])
			Key_Event(i, false, 0);
		keydown[i] = 0;
		key_repeats[i] = 0;
	}
}


/*
===================
Key_GetKey
===================
*/
int Key_GetKey (void)
{
	key_waiting = -1;

	while (key_waiting == -1)
		Sys_SendKeyEvents ();

	return key_waiting;
}
