//-----------------------------------------------------------------------------
//
//  $Logfile:: /Source/gl_light.c                                             $
// $Revision:: 4                                                              $
//   $Author:: Burgermac                                                      $
//     $Date:: 8/8/01 4:46 PM                                                 $
//
// Copyright (C) 1998 by Ritual Entertainment, Inc.
// All rights reserved.
//
// This source is may not be distributed and/or modified without
// expressly written permission by Ritual Entertainment, Inc.
//
// DESCRIPTION:
//

#include "gl_local.h"

int	r_dlightframecount;

void R_SetCacheState( msurface_t *surf );
extern void R_BuildLightMap (msurface_t *surf, byte *dest, int stride);

//#define	DLIGHT_CUTOFF	8	/* FS: Now a CVAR */

/*
=============================================================================

DYNAMIC LIGHTS BLEND RENDERING

=============================================================================
*/

static void R_RenderDlight (dlight_t *light)
{
	int		i, j;
	float	a;
	vec3_t	v;
	float	rad;

	/// Don't support neg light
	if (( light->color[0] < 0 )
		|| ( light->color[1] < 0 )
		|| ( light->color[2] < 0 ))
		return;

	rad = light->intensity * 0.35f;

	VectorSubtract (light->origin, r_origin, v);
#if 0
	// FIXME?
	if (VectorLength (v) < rad)
	{	// view is inside the dlight
		V_AddBlend (light->color[0], light->color[1], light->color[2], light->intensity * 0.0003, v_blend);
		return;
	}
#endif

	qglBegin (GL_TRIANGLE_FAN);
	qglColor3f (light->color[0]*0.2f, light->color[1]*0.2f, light->color[2]*0.2f);
	for (i=0 ; i<3 ; i++)
		v[i] = light->origin[i] - vpn[i]*rad;
	qglVertex3fv (v);
	qglColor3f (0,0,0);
	for (i=16 ; i>=0 ; i--)
	{
		a = i/16.0f * M_PI*2;
		for (j=0 ; j<3 ; j++)
			v[j] = light->origin[j] + vright[j]*cos(a)*rad
				+ vup[j]*sin(a)*rad;
		qglVertex3fv (v);
	}
	qglEnd ();
}

/*
=============
R_RenderDlights
=============
*/
void R_RenderDlights (void)
{
	int		i;
	dlight_t	*l;

	if (!gl_flashblend->value)
		return;

	r_dlightframecount = r_framecount + 1;	// because the count hasn't
											//  advanced yet for this frame
	qglDepthMask (0);
	qglDisable (GL_TEXTURE_2D);
	qglShadeModel (GL_SMOOTH);
	qglEnable (GL_BLEND);
	qglBlendFunc (GL_ONE, GL_ONE);

	l = r_newrefdef.dlights;
	for (i=0 ; i<r_newrefdef.num_dlights ; i++, l++)
		R_RenderDlight (l);

	qglColor3f (1,1,1);
	qglDisable (GL_BLEND);
	qglEnable (GL_TEXTURE_2D);
	qglBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	qglDepthMask (1);
}


/*
=============================================================================

DYNAMIC LIGHTS

=============================================================================
*/

/*
=============
R_MarkLights
=============
*/
#ifndef AMD3D

void R_MarkLights (dlight_t *light, int bit, mnode_t *node)
{
	cplane_t	*splitplane;
	float		dist;
	msurface_t	*surf;
	int			i;
	vec3_t		offset;
	vec3_t		ofs;
	float		mat[3][3];

	if (node->contents != -1)
		return;

	splitplane = node->plane;

	if ( currententity->angles[0] || currententity->angles[1] || currententity->angles[2])
	{
		TransposeMatrix( currententity->mat, mat );
		VectorSubtract( light->origin, currententity->origin, ofs );
		MatrixTransformVector( ofs, mat, offset );
	}
	else
	{
		VectorSubtract( light->origin, currententity->origin, offset );
	}
	dist = DotProduct ( offset, splitplane->normal) - splitplane->dist;

	if (dist > light->intensity-gl_lightcutoff->value) /* FS: Was DLIGHT_CUTOFF */
	{
		R_MarkLights (light, bit, node->children[0]);
		return;
	}
	if (dist < -light->intensity+gl_lightcutoff->value) /* FS: Was DLIGHT_CUTOFF */
	{
		R_MarkLights (light, bit, node->children[1]);
		return;
	}

// mark the polygons
	surf = r_worldmodel->surfaces + node->firstsurface;
	for (i=0 ; i<node->numsurfaces ; i++, surf++)
	{
		if (surf->dlightframe != r_dlightframecount)
		{
			surf->dlightbits = 0;
			surf->dlightframe = r_dlightframecount;
		}
		surf->dlightbits |= bit;
	}

	R_MarkLights (light, bit, node->children[0]);
	R_MarkLights (light, bit, node->children[1]);
}

#endif

/*
=============
R_PushDlights
=============
*/
void R_PushDlights (void)
{
	int		i;
	dlight_t	*l;

	if (gl_flashblend->value)
		return;

	r_dlightframecount = r_framecount + 1;	// because the count hasn't
											//  advanced yet for this frame
	l = r_newrefdef.dlights;

	for (i=0 ; i<r_newrefdef.num_dlights ; i++, l++)
		R_MarkLights ( l, 1<<i, r_worldmodel->nodes );
}


/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

vec3_t			pointcolor;
cplane_t		*lightplane;		// used as shadow plane
vec3_t			lightspot;

#ifndef AMD3D

static int RecursiveLightPoint (mnode_t *node, vec3_t start, vec3_t end)
{
	float		front, back, frac;
	int			side;
	cplane_t	*plane;
	vec3_t		mid;
	msurface_t	*surf;
	int			s, t, ds, dt;
	int			i;
	mtexinfo_t	*tex;
	byte		*lightmap;
	int			maps;
	int			r;

	if (node->contents != -1)
		return -1;		// didn't hit anything

// calculate mid point

// FIXME: optimize for axial
	plane = node->plane;
	front = DotProduct (start, plane->normal) - plane->dist;
	back = DotProduct (end, plane->normal) - plane->dist;
	side = front < 0;

	if ( (back < 0) == side)
		return RecursiveLightPoint (node->children[side], start, end);

	frac = front / (front-back);
	mid[0] = start[0] + (end[0] - start[0])*frac;
	mid[1] = start[1] + (end[1] - start[1])*frac;
	mid[2] = start[2] + (end[2] - start[2])*frac;

// go down front side
	r = RecursiveLightPoint (node->children[side], start, mid);
	if (r >= 0)
		return r;		// hit something

	if ( (back < 0) == side )
		return -1;		// didn't hit anuthing

// check for impact on this node
	VectorCopy (mid, lightspot);
	lightplane = plane;

	surf = r_worldmodel->surfaces + node->firstsurface;
	for (i=0 ; i<node->numsurfaces ; i++, surf++)
	{
		if (surf->flags&(SURF_DRAWTURB|SURF_DRAWSKY))
			continue;	// no lightmaps

		tex = surf->texinfo;

		s = DotProduct (mid, tex->vecs[0]) + tex->vecs[0][3];
		t = DotProduct (mid, tex->vecs[1]) + tex->vecs[1][3];;

		if (s < surf->texturemins[0] ||
		t < surf->texturemins[1])
			continue;

		ds = s - surf->texturemins[0];
		dt = t - surf->texturemins[1];

		if ( ds > surf->extents[0] || dt > surf->extents[1] )
			continue;

		if (surf->flags&SURF_DRAWNONLIT)
		{
			if (surf->texinfo->flags & SURF_USECOLOR)
			{
				pointcolor[0] = tex->surf->nonlit * tex->surf->color[0];
				pointcolor[1] = tex->surf->nonlit * tex->surf->color[1];
				pointcolor[2] = tex->surf->nonlit * tex->surf->color[2];
			}
			else
			{
				pointcolor[0] = tex->surf->nonlit;
				pointcolor[1] = tex->surf->nonlit;
				pointcolor[2] = tex->surf->nonlit;
			}
			return 1;
		}

		if (!surf->samples)
			return 0;

		ds >>= 4;
		dt >>= 4;

		lightmap = surf->samples;
		VectorCopy (vec3_origin, pointcolor);
		if (lightmap && r_newrefdef.lightstyles)
		{
			vec3_t scale;

			lightmap += 3*(dt * ((surf->extents[0]>>4)+1) + ds);

			for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ;
					maps++)
			{
				for (i=0 ; i<3 ; i++)
					scale[i] = gl_modulate->value*r_newrefdef.lightstyles[surf->styles[maps]].rgb[i];

				pointcolor[0] += lightmap[0] * scale[0] * (1.0/255);
				pointcolor[1] += lightmap[1] * scale[1] * (1.0/255);
				pointcolor[2] += lightmap[2] * scale[2] * (1.0/255);
				lightmap += 3*((surf->extents[0]>>4)+1) *
						((surf->extents[1]>>4)+1);
			}
		}

		return 1;
	}

// go down back side
	return RecursiveLightPoint (node->children[!side], mid, end);
}
#endif

void R_GetSurfaceColorAtST( msurface_t *surf,float ss,float tt,vec3_t color )
{
	int i;
	int s, t, ds, dt;
	byte *lightmap;
	int maps;

	if (surf->flags&(SURF_DRAWTURB|SURF_DRAWSKY) || (!surf->samples) )
	{
		color[0] = 1.0;
		color[1] = 1.0;
		color[2] = 1.0;
		return;
	}
	s =  ss * surf->texinfo->image->width;
	t =  tt * surf->texinfo->image->height;


	if ( s < surf->texturemins[0] )
	{
		s = surf->texturemins[0];
	}

	if ( t < surf->texturemins[1] )
	{
		t = surf->texturemins[1];
	}

	ds = s - surf->texturemins[0];
	dt = t - surf->texturemins[1];

	if ( ds > surf->extents[0] )
	{
		ds = surf->extents[ 0 ];
	}

	if ( dt > surf->extents[1] )
	{
		dt = surf->extents[ 1 ];
	}

	ds >>= 4;
	dt >>= 4;


	lightmap = surf->samples;
	VectorCopy ( vec3_origin, color );
	if (lightmap)
	{
		vec3_t scale;

		lightmap += 3*(dt * ((surf->extents[0]>>4)+1) + ds);

		for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255;
		     maps++)
		{
			if ( r_newrefdef.lightstyles )
			{
				for (i=0; i<3; i++)
					scale[i] = gl_modulate->value*r_newrefdef.lightstyles[surf->styles[maps]].rgb[i];
			}
			else
			{
				for (i=0; i<3; i++)
					scale[i] = gl_modulate->value;
			}
			color[0] += lightmap[0] * scale[0] * (1.0/255);
			color[1] += lightmap[1] * scale[1] * (1.0/255);
			color[2] += lightmap[2] * scale[2] * (1.0/255);
			lightmap += 3*((surf->extents[0]>>4)+1) *
			            ((surf->extents[1]>>4)+1);
		}
	}
}

/*
===============
R_LightPoint
===============
*/
void R_LightPoint ( vec3_t p, vec3_t color, qboolean dlight)
{
	vec3_t end;
	float r;
	int lnum;
	dlight_t *dl;
	vec3_t dist;
	float add;
	float mat[3][3];
	qboolean trans;
	vec3_t ofs;

	if (!r_worldmodel->lightdata)
	{
		color[0] = color[1] = color[2] = 1.0;
		return;
	}

	end[0] = p[0];
	end[1] = p[1];
	end[2] = p[2] - 2048;

	r = RecursiveLightPoint (r_worldmodel->nodes, p, end);

	if (r == -1)
	{
		color[0] = color[1] = color[2] = 1.0;
	}
	else
	{
		VectorCopy (pointcolor, color);
	}

	if ( dlight )
	{
		//
		// add dynamic lights
		//
		dl = r_newrefdef.dlights;
		TransposeMatrix( currententity->mat, mat );
		trans = ( currententity->angles[0] || currententity->angles[1] || currententity->angles[2] );
		for (lnum=0 ; lnum<r_newrefdef.num_dlights ; lnum++, dl++)
		{
			if ( trans )
			{
				VectorSubtract( dl->origin, currententity->origin, ofs );
				MatrixTransformVector( ofs, mat, dist );
			}
			else
			{
				VectorSubtract( currententity->origin, dl->origin, dist );
			}
			add = dl->intensity - VectorLength(dist);
			add *= (1.0/256);
			if (add > 0)
			{
				VectorMA (color, add, dl->color, color);
			}
		}
	}

	VectorScale (color, gl_modulate->value, color);
}

/*
=============================================================================

LIGHT BURNING

=============================================================================
*/

static int RecursiveLightBurn (mnode_t *node, vec3_t start, vec3_t end, int burnvalue)
{
	int r;
	float front, back, frac;
	int side;
	cplane_t *plane;
	vec3_t mid;
	msurface_t *surf;
	int s, t, ds, dt, dds, ddt;
	int i;
	mtexinfo_t *tex;
	byte *lightmap;
	int maps;
	float ul,ur,ll,lr;

	if (node->contents != -1)
		return -1;      // didn't hit anything

// calculate mid point

// FIXME: optimize for axial
	plane = node->plane;
	front = DotProduct ( start, plane->normal ) - plane->dist;
	back = DotProduct ( end, plane->normal ) - plane->dist;
	side = front < 0;

	if ( (back < 0) == side)
		return RecursiveLightBurn ( node->children[side], start, end, burnvalue );

	frac = front / (front-back);
	mid[0] = start[0] + (end[0] - start[0])*frac;
	mid[1] = start[1] + (end[1] - start[1])*frac;
	mid[2] = start[2] + (end[2] - start[2])*frac;

// go down front side
	r = RecursiveLightBurn ( node->children[side], start, mid, burnvalue );
	if (r >= 0)
		return r;       // hit something

	if ( (back < 0) == side )
		return -1;      // didn't hit anuthing

// check for impact on this node
	VectorCopy ( mid, lightspot );
	lightplane = plane;

	surf = r_worldmodel->surfaces + node->firstsurface;
	for (i=0; i<node->numsurfaces; i++, surf++)
	{
		if (surf->flags&(SURF_DRAWTURB|SURF_DRAWSKY|SURF_DRAWNONLIT))
			continue;   // no lightmaps

		tex = surf->texinfo;

		s = DotProduct ( mid, tex->vecs[0] ) + tex->vecs[0][3];
		t = DotProduct ( mid, tex->vecs[1] ) + tex->vecs[1][3];

		if (s < surf->texturemins[0] ||
		    t < surf->texturemins[1])
			continue;

		ds = s - surf->texturemins[0];
		dt = t - surf->texturemins[1];

		if ( ds > surf->extents[0] || dt > surf->extents[1] )
			continue;

		if (!surf->samples)
			return 0;

		surf->dlightframe = -89324;

		s = ds >> 4;
		t = dt >> 4;

		dds = ds - (s<<4);
		ddt = dt - (t<<4);
		ul = 1.0f - (sqrt((double)(dds*dds + ddt*ddt))/11.32f);
		if (ul<0) ul = 0;

		ddt = ((t+1)<<4) - dt;
		ll = 1.0f - (sqrt((double)(dds*dds + ddt*ddt))/11.32f);
		if (ll<0) ll = 0;

		dds = ((s+1)<<4) - ds;
		ddt = dt - (t<<4);
		ur = 1.0f - (sqrt((double)(dds*dds + ddt*ddt))/11.32f);
		if (ur<0) ur = 0;

		ddt = ((t+1)<<4) - dt;
		lr = 1.0f - (sqrt((double)(dds*dds + ddt*ddt))/11.32f);
		if (lr<0) lr = 0;

		lightmap = surf->samples;
		r = 1;
		if (lightmap)
		{
			byte * lmap;
			float value;

			lightmap += 3*(t * ((surf->extents[0]>>4)+1) + s);

			for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255;
			     maps++)
			{
				lmap = lightmap;
				for ( i = 0; i < 3; i++, lmap++ )
				{
					value = *lmap + burnvalue*ul;
					if (value<0) value = 0;
					if (value>255) value = 255;
					*lmap = value;
				}
				if (s <= surf->extents[0])
				{
					lmap = lightmap+3;
					for ( i = 0; i < 3; i++, lmap++ )
					{
						value = *lmap + burnvalue*ur;
						if (value<0) value = 0;
						if (value>255) value = 255;
						*lmap = value;
					}
				}
				if (t <= surf->extents[1])
				{
					lmap = lightmap+(((surf->extents[0]>>4)+1)*3);
					for ( i = 0; i < 3; i++, lmap++ )
					{
						value = *lmap + burnvalue*ll;
						if (value<0) value = 0;
						if (value>255) value = 255;
						*lmap = value;
					}
					if (s <= surf->extents[0])
					{
						lmap = lightmap+(((surf->extents[0]>>4)+2)*3);
						for ( i = 0; i < 3; i++, lmap++ )
						{
							value = *lmap + burnvalue*lr;
							if (value<0) value = 0;
							if (value>255) value = 255;
							*lmap = value;
						}
					}
				}
				lightmap += 3*((surf->extents[0]>>4)+1) *
				            ((surf->extents[1]>>4)+1);
			}
		}
		GL_UpdateLightmap( surf );
		return 1;
	}
// go down back side
	return RecursiveLightBurn ( node->children[!side], mid, end, burnvalue );
}

void R_BurnPoint (vec3_t p, vec3_t dir, int value)
{
	vec3_t start;
	vec3_t end;
	vec3_t u,v;
	vec3_t xoff, yoff;

	PerpendicularVector( u, dir );
	CrossProduct( u, dir, v );

	VectorScale( u, 8, xoff );
	VectorScale( v, 8, yoff );
	VectorCopy( p, start );

	end[0] = start[0] + dir[0]*2048;
	end[1] = start[1] + dir[1]*2048;
	end[2] = start[2] + dir[2]*2048;

	RecursiveLightBurn ( r_worldmodel->nodes, start, end, value/5 );

	VectorAdd( start, xoff, start );
	VectorAdd( end, xoff, end );
	RecursiveLightBurn ( r_worldmodel->nodes, start, end, value/5 );

	VectorAdd( start, yoff, start );
	VectorAdd( end, yoff, end );
	RecursiveLightBurn ( r_worldmodel->nodes, start, end, value/5 );

	VectorSubtract( start, xoff, start );
	VectorSubtract( end, xoff, end );
	VectorSubtract( start, xoff, start );
	VectorSubtract( end, xoff, end );
	RecursiveLightBurn ( r_worldmodel->nodes, start, end, value/5 );

	VectorSubtract( start, yoff, start );
	VectorSubtract( end, yoff, end );
	VectorSubtract( start, yoff, start );
	VectorSubtract( end, yoff, end );
	RecursiveLightBurn ( r_worldmodel->nodes, start, end, value/5 );
}

//===================================================================

float	s_blocklights[34*34*3];
/*
===============
R_AddDynamicLights
===============
*/
static void R_AddDynamicLights (msurface_t *surf)
{
	int			lnum;
	int			sd, td;
	float			fdist, frad, fminlight;
	vec3_t		impact, local;
	int			s, t;
	int			i;
	int			smax, tmax;
	mtexinfo_t	*tex;
	dlight_t		*dl;
	float		*pfBL;
	float		fsacc, ftacc;
	float		scalefactor;
	vec3_t		offset;
	vec3_t		ofs;
	float		mat[3][3];
	qboolean	trans;
	float		tx_length, ty_length;

	TransposeMatrix( currententity->mat, mat );
	trans = ( currententity->angles[0] || currententity->angles[1] || currententity->angles[2] );

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	tex = surf->texinfo;

	for (lnum=0 ; lnum<r_newrefdef.num_dlights ; lnum++)
	{
		if ( !(surf->dlightbits & (1<<lnum) ) )
			continue;		// not lit by this light

		dl = &r_newrefdef.dlights[lnum];
		frad = dl->intensity;

		if ( trans )
		{
			VectorSubtract( dl->origin, currententity->origin, ofs );
			MatrixTransformVector( ofs, mat, offset );
		}
		else
		{
			VectorSubtract( dl->origin, currententity->origin, offset );
		}
		fdist = DotProduct ( offset, surf->plane->normal ) - surf->plane->dist;

		frad -= fabs(fdist);
		// rad is now the highest intensity on the plane

		fminlight = gl_lightcutoff->value; /* FS: Was DLIGHT_CUTOFF */
		if (frad < fminlight)
			continue;
		fminlight = frad - fminlight;
		tx_length = VectorLength( tex->vecs[0] );
		ty_length = VectorLength( tex->vecs[1] );
		if ( tx_length < ty_length )
			scalefactor = tx_length;
		else
			scalefactor = ty_length;
		fminlight *= scalefactor;
		scalefactor = 1.0f / scalefactor;

		for (i=0 ; i<3 ; i++)
		{
			impact[i] = offset[i] -
					surf->plane->normal[i]*fdist;
		}

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];

		pfBL = s_blocklights;
		for (t = 0, ftacc = 0 ; t<tmax ; t++, ftacc += 16)
		{
			td = local[1] - ftacc;
			if ( td < 0 )
				td = -td;

			for ( s=0, fsacc = 0 ; s<smax ; s++, fsacc += 16, pfBL += 3)
			{
				sd = Q_ftol( local[0] - fsacc );

				if ( sd < 0 )
					sd = -sd;

				if (sd > td)
					fdist = sd + (td>>1);
				else
					fdist = td + (sd>>1);

				if ( fdist < fminlight )
				{
					if ( scalefactor != 1 )
						fdist *= scalefactor;
					pfBL[0] += ( frad - fdist ) * dl->color[0];
					pfBL[1] += ( frad - fdist ) * dl->color[1];
					pfBL[2] += ( frad - fdist ) * dl->color[2];
				}
			}
		}
	}
}


/*
** R_SetCacheState
*/
void R_SetCacheState( msurface_t *surf )
{
	int maps;

	for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ;
		 maps++)
	{
		surf->cached_light[maps] = r_newrefdef.lightstyles[surf->styles[maps]].rgbvalue;
	}
/* Knightmare- added for lightmap update batching */
#ifdef BATCH_LM_UPDATES
	// mark if dynamicly lit
	surf->cached_dlight = (surf->dlightframe == r_framecount);
#endif
}

/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the floating format in blocklights
===============
*/
void R_BuildLightMap (msurface_t *surf, byte *dest, int stride)
{
	int			smax, tmax;
	int			r, g, b, a, max;
	int			i, j, size;
	byte		*lightmap;
	float		scale[4];
	int			nummaps;
	float		*bl;
	int monolightmap;

	if (surf->texinfo->flags & (SURF_SKY|SURF_WARP|SURF_WAVY) || surf->texinfo->surf->translucence > 0)
		ri.Sys_Error (ERR_DROP, "R_BuildLightMap called for non-lit surface");

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	size = smax*tmax;
	if (size > (sizeof(s_blocklights)>>4))
		ri.Sys_Error (ERR_DROP, "Bad blocklights size");

// set to full bright if no light data
	if (!surf->samples)
	{
		for (i = 0; i < size * 3; i++)
			s_blocklights[i] = 255;
		goto store;
	}

	// count the # of maps
	for ( nummaps = 0 ; nummaps < MAXLIGHTMAPS && surf->styles[nummaps] != 255 ;
		 nummaps++)
		;

	lightmap = surf->samples;

	// add all the lightmaps
	if ( nummaps == 1 )
	{
		int maps;

		for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ;
			 maps++)
		{
			bl = s_blocklights;

			for (i=0 ; i<3 ; i++)
				scale[i] = gl_modulate->value*r_newrefdef.lightstyles[surf->styles[maps]].rgb[i];

			if ( scale[0] == 1.0F &&
				 scale[1] == 1.0F &&
				 scale[2] == 1.0F )
			{
				for (i=0 ; i<size ; i++, bl+=3)
				{
					bl[0] = lightmap[i*3+0];
					bl[1] = lightmap[i*3+1];
					bl[2] = lightmap[i*3+2];
				}
			}
			else
			{
				for (i=0 ; i<size ; i++, bl+=3)
				{
					bl[0] = lightmap[i*3+0] * scale[0];
					bl[1] = lightmap[i*3+1] * scale[1];
					bl[2] = lightmap[i*3+2] * scale[2];
				}
			}
			lightmap += size*3;		// skip to next lightmap
		}
	}
	else
	{
		int maps;

		memset( s_blocklights, 0, sizeof( s_blocklights[0] ) * size * 3 );

		for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ;
			 maps++)
		{
			bl = s_blocklights;

			for (i=0 ; i<3 ; i++)
				scale[i] = gl_modulate->value*r_newrefdef.lightstyles[surf->styles[maps]].rgb[i];

			if ( scale[0] == 1.0F &&
				 scale[1] == 1.0F &&
				 scale[2] == 1.0F )
			{
				for (i=0 ; i<size ; i++, bl+=3 )
				{
					bl[0] += lightmap[i*3+0];
					bl[1] += lightmap[i*3+1];
					bl[2] += lightmap[i*3+2];
				}
			}
			else
			{
				for (i=0 ; i<size ; i++, bl+=3)
				{
					bl[0] += lightmap[i*3+0] * scale[0];
					bl[1] += lightmap[i*3+1] * scale[1];
					bl[2] += lightmap[i*3+2] * scale[2];
				}
			}
			lightmap += size*3;		// skip to next lightmap
		}
	}

// add all the dynamic lights
	if (surf->dlightframe == r_framecount)
		R_AddDynamicLights (surf);

// put into texture format
store:
	stride -= (smax<<2);
	bl = s_blocklights;

	monolightmap = gl_monolightmap->string[0];

	if ( monolightmap == '0' )
	{
		for (i=0 ; i<tmax ; i++, dest += stride)
		{
			for (j=0 ; j<smax ; j++)
			{

				r = Q_ftol( bl[0] );
				g = Q_ftol( bl[1] );
				b = Q_ftol( bl[2] );

				// catch negative lights
				if (r < 0)
					r = 0;
				if (g < 0)
					g = 0;
				if (b < 0)
					b = 0;

				/*
				** determine the brightest of the three color components
				*/
				if (r > g)
					max = r;
				else
					max = g;
				if (b > max)
					max = b;

				/*
				** alpha is ONLY used for the mono lightmap case.  For this reason
				** we set it to the brightest of the color components so that
				** things don't get too dim.
				*/
				a = max;

				/*
				** rescale all the color components if the intensity of the greatest
				** channel exceeds 1.0
				*/
				if (max > 255)
				{
					float t = 255.0F / max;

					r = r*t;
					g = g*t;
					b = b*t;
					a = a*t;
				}

				dest[0] = r;
				dest[1] = g;
				dest[2] = b;
				dest[3] = a;

				bl += 3;
				dest += 4;
			}
		}
	}
	else
	{
		for (i=0 ; i<tmax ; i++, dest += stride)
		{
			for (j=0 ; j<smax ; j++)
			{

				r = Q_ftol( bl[0] );
				g = Q_ftol( bl[1] );
				b = Q_ftol( bl[2] );

				// catch negative lights
				if (r < 0)
					r = 0;
				if (g < 0)
					g = 0;
				if (b < 0)
					b = 0;

				/*
				** determine the brightest of the three color components
				*/
				if (r > g)
					max = r;
				else
					max = g;
				if (b > max)
					max = b;

				/*
				** alpha is ONLY used for the mono lightmap case.  For this reason
				** we set it to the brightest of the color components so that
				** things don't get too dim.
				*/
				a = max;

				/*
				** rescale all the color components if the intensity of the greatest
				** channel exceeds 1.0
				*/
				if (max > 255)
				{
					float t = 255.0F / max;

					r = r*t;
					g = g*t;
					b = b*t;
					a = a*t;
				}

				/*
				** So if we are doing alpha lightmaps we need to set the R, G, and B
				** components to 0 and we need to set alpha to 1-alpha.
				*/
				switch ( monolightmap )
				{
				case 'L':
				case 'I':
					r = a;
					g = b = 0;
					break;
				case 'C':
					// try faking colored lighting
					a = 255 - ((r+g+b)/3);
					r *= a/255.0;
					g *= a/255.0;
					b *= a/255.0;
					break;
				case 'A':
				default:
					r = g = b = 0;
					a = 255 - a;
					break;
				}

				dest[0] = r;
				dest[1] = g;
				dest[2] = b;
				dest[3] = a;

				bl += 3;
				dest += 4;
			}
		}
	}
}
